#include "AD7768.h"

AD7768::AD7768(byte CSpin, byte STARTpin, byte DCLKpin, byte DRDYpin, byte D0pin, byte D1pin, byte D2pin, byte D3pin, byte D4pin, byte D5pin, byte D6pin, byte D7pin):
      _CSpin(CSpin),
      _STARTpin(STARTpin),
      _VCM(0),
      _DUEScaler(1),
      _AD7768SPISettings(SPISettings(10000000, MSBFIRST, SPI_MODE0)) {
        // Create an array for the pin numbers to speed up access.
        _dataPins[0] = DCLKpin;
        _dataPins[1] = DRDYpin;
        _dataPins[2] = D0pin;
        _dataPins[3] = D1pin;
        _dataPins[4] = D2pin;
        _dataPins[5] = D3pin;
        _dataPins[6] = D4pin;
        _dataPins[7] = D5pin;
        _dataPins[8] = D6pin;
        _dataPins[9] = D7pin;
        }

void AD7768::begin(void) {
  // Configure Pins.
  pinMode(_CSpin, OUTPUT);
  pinMode(_STARTpin, OUTPUT);

  for (byte i = 0; i<10; i++) {
    pinMode(_dataPins[i], INPUT);
  }

  for (byte i = 0; i<8; i++) {
    _voltages[i] = 0;
    _rawData[i] = 0;
  }

  // Set initial pin states.
  digitalWrite(_CSpin, HIGH);
  digitalWrite(_STARTpin, HIGH);

  // Initialize device configuration.
  delay(2); // As per table 1 on AD7768, wait for start-up time before using SPI (>1.66ms).

  // Configure for all channels enabled.
  std::vector<byte> data;
  data.push_back(AD7768_WRITE | AD7768_CHSTDB);
  data.push_back(0b00000000); // Value (0 = channel enabled)
  SendCommand(data);
  data.clear();

  // Channel Mode A
  data.push_back(AD7768_WRITE | AD7768_CHMODEA);
  data.push_back(0b00001000); // Sinc5 Filter, x32 decimation
  SendCommand(data);
  data.clear();

  // Channel Mode Select
  data.push_back(AD7768_WRITE | AD7768_CHMODSEL);
  data.push_back(0b00000000); // All channels set to Mode A.
  SendCommand(data);
  data.clear();

  // Power Mode Select
  data.push_back(AD7768_WRITE | AD7768_POWERMODE);
  data.push_back(0b00110011); // Fast Mode (Fmod = MCLK/4), LVDS disabled, MCLK/4.
  SendCommand(data);
  data.clear();

  // General Device Configuration
  data.push_back(AD7768_WRITE | AD7768_GENCONF);
  data.push_back(0b00000001); // SYNC_OUT normal timing, VCM enabled, VCM = 1.65V.
  SendCommand(data);
  data.clear();

  // Data Control
  data.push_back(AD7768_WRITE | AD7768_DATACONTROL);
  data.push_back(0b00000000); // Continuous mode, SPI_SYNC low, no soft reset this time.
  SendCommand(data);
  data.clear();

  // Interface Control
  data.push_back(AD7768_WRITE | AD7768_INTCONF);
  data.push_back(0b00000001); // No CRC, DCLK = MCLK/4.
  SendCommand(data);
  data.clear();

  // Chopping Control Register
  data.push_back(AD7768_WRITE | AD7768_CHOPCTL);
  data.push_back(0b00001010); // Chop at fmod/32.
  SendCommand(data);
  
  // Load New Configuration by pulsing Start signal.
  PulseStart();
}

void AD7768::SendCommand(std::vector<byte> &data) {
  // Configure the SPI Port for this chip.
  SPI.beginTransaction(_AD7768SPISettings);
  
  // Activate CS pin.
  digitalWrite(_CSpin, LOW);
  
  // Write the data out.
  for (std::vector<byte>::iterator i=data.begin(); i != data.end(); ++i) {
    // For this we will ignore the returned data.
    SPI.transfer(*i);
  }
  
  // Deactivate CS pin.
  digitalWrite(_CSpin, HIGH);
  SPI.endTransaction();
}

void AD7768::PulseStart(void) {
  // On the AD7768, you must pulse the START pin (possibly twice) after changing configuration.
  digitalWrite(_STARTpin, LOW);
  delayMicroseconds(1); // Hold time for START pulse is must be greater than MCLK (~30ns).
  digitalWrite(_STARTpin, HIGH);
  delayMicroseconds(1);
  digitalWrite(_STARTpin, LOW);
  delayMicroseconds(1); 
  digitalWrite(_STARTpin, HIGH);
}

byte* AD7768::getDataPins(void) {
  return _dataPins;
}

// This function converts the twos complement formatted 32-bit values into floating point voltages for storage.
void AD7768::writeValue(unsigned int* values) {  
  for (byte i=0; i<8; i++) {
    signed long signedvalue;
    unsigned long datavalue = values[i] & 0xFFFFFF; // Strip the 8-bit header from the 32-bit data packet.
    
    if (datavalue & 0x800000) { // Check the sign bit. Note that it is in the D23 position instead of D31 as for a normal signed long.
      signedvalue = (signed long)(datavalue & 0x7FFFFF) - 0x800000;
    }
    else {
      signedvalue = (signed long)datavalue;
    }

    _rawData[i] = (unsigned int)(signedvalue + 4613734); // This should be the code for 1.65V on our scale to cancel the VCM effect.
    _voltages[i] = ((float)signedvalue/(float)0xFFFFFF) * 6;
  }
}

float AD7768::getValue(byte channel) {
  if (channel >= 0 and channel <= 7) {
    return _voltages[channel];
  }
  else return -1;
}

unsigned int AD7768::getData(byte channel) {
  if (channel >= 0 and channel <= 7) {
    return _rawData[channel];
  }
  else return 0;
}

float AD7768::getVCM(void) {
  return _VCM;
}

void AD7768::calibrate(void) {
}

