/**********************************************************************************************
 * 
 * 24 Channel Monitor Firmware
 * Version 2.1 (X3 Hardware Revision)
 * 2018-03-13
 * Brian Neltner, Neltner Labs (neltnerb@neltnerlabs.com)
 * 
 * Description: This firmware is a refactoring of the old codebase while adding the new features
 * to support the new X3 version of the stimulator board. This firmware is for the Teensy 3.5 used
 * as a signal monitor which streams ADC data to the LabView systems continuously using the COBS
 * serial packet format.
 * 
 * Link to Neltner Lab's serial receive library:
 * https://github.com/neltnerb/Teensy-Reactor-Controller
 * for documentation in using the serial communications library.
 * 
 **********************************************************************************************/

#include <Arduino.h>
#include <vector>
#include "SerialEval.h"
#include "ADC.h"

// Firmware defines.
#define Title "Brain Stimulator Monitor"
#define FirmwareVersion "2.7"
#define BoardVersion "12"

ADC *adc = new ADC();

// Global variables.
boolean debugSerial = false; // Set to false to disable sending any responses to commands over the USB interface.
boolean streaming = false; // This flag tells the system if it is currently streaming data.

// These are the prototypes for each individual serial command handler.

serialcommand(getVersion);
serialcommand(startStreaming);
serialcommand(stopStreaming);
serialcommand(adcRead);

void setup() {
  Serial.begin(115200); // Initialize the USB interface.
  delay(1000);
 
  pinMode(A1, INPUT);
  pinMode(A2, INPUT);
  pinMode(A3, INPUT);
  pinMode(A4, INPUT);
  pinMode(A5, INPUT);
  pinMode(A6, INPUT);
  pinMode(A7, INPUT);
  pinMode(A8, INPUT);

  // Internal buffer for averaging. For maximum speed set to 1 or 0.
  
  adc->setAveraging(1, ADC_0);
  adc->setAveraging(1, ADC_1);

  // Resolution. 16-bit resolution will confuse LabView, so stick with 15 or less.
  
  adc->setResolution(13, ADC_0);
  adc->setResolution(13, ADC_1);

  // Configure to use the 3.3V onboard regulator as a voltage reference. Use
  // an external reference or the internal more precise 1.2V regulator for improved
  // performance.
  
  adc->setReference(ADC_REFERENCE::REF_3V3, ADC_0);
  adc->setReference(ADC_REFERENCE::REF_3V3, ADC_1);

  // Set the conversion speed. Conversion is how long the ADC measures the internal
  // buffering capacitor.

  adc->setConversionSpeed(ADC_CONVERSION_SPEED::VERY_HIGH_SPEED, ADC_0);
  adc->setConversionSpeed(ADC_CONVERSION_SPEED::VERY_HIGH_SPEED, ADC_1);

  // Set the sampling speed. Sampling speed is how long the internal buffering capacitor
  // is connected to the specified multiplexed input pin prior to conversion.

  adc->setSamplingSpeed(ADC_SAMPLING_SPEED::VERY_HIGH_SPEED, ADC_0);
  adc->setSamplingSpeed(ADC_SAMPLING_SPEED::VERY_HIGH_SPEED, ADC_1);

  // Register serial commands to the serial library.
  
  registerCommand("GetVersion", getVersion, "Get the Device ID and Firmware Version.");
  registerCommand("StartStreaming", startStreaming, "Begin Continuous Streaming Mode in COBS format.");
  registerCommand("StopStreaming", stopStreaming, "End Continuous Streaming Mode.");
  registerCommand("ADCRead", adcRead, "Collect 8000 datapoints and send in COBS format.");
  
  Serial.println("System Started.");
}

void loop() {
  checkSerial(); // Check for any new received commands on the USB interface.
  
  if (streaming) {
    // Get the time.
    unsigned long currenttime = micros();
    
    // Get the analog input values.
    unsigned int analog[8];
    ADC::Sync_result result = adc->analogSyncRead(A1, A2);
    analog[6] = result.result_adc0;
    analog[7] = result.result_adc1;
    result = adc->analogSyncRead(A4, A3);
    analog[5] = result.result_adc0;
    analog[4] = result.result_adc1;
    analog[2] = adc->analogRead(A5);
    analog[3] = adc->analogRead(A6);
    analog[0] = adc->analogRead(A7);
    analog[1] = adc->analogRead(A8);
    
    byte packet[22]; // 16 bytes of data, four bytes for timestamp, two bytes of overhead in COBS format.

    packet[0] = currenttime>>24;
    packet[1] = currenttime>>16;
    packet[2] = currenttime>>8;
    packet[3] = currenttime;
    packet[4] = analog[0]>>8;
    packet[5] = analog[0];
    packet[6] = analog[1]>>8;
    packet[7] = analog[1];
    packet[8] = analog[2]>>8;
    packet[9] = analog[2];
    packet[10] = analog[3]>>8;
    packet[11] = analog[3];
    packet[12] = analog[4]>>8;
    packet[13] = analog[4];
    packet[14] = analog[5]>>8;
    packet[15] = analog[5];
    packet[16] = analog[6]>>8;
    packet[17] = analog[6];
    packet[18] = analog[7]>>8;
    packet[19] = analog[7];

    // Convert to COBS format.
    byte data[22];
    data[0] = 21; // Default to the end of the packet.
    data[21] = 0x00; // This is always set to the termination character.
    byte lastzero = 0;
    
    for (byte i=0; i<20; i++) {
      if (packet[i] == 0x00) {
          data[lastzero] = (i+1)-lastzero; // Change the last offset byte to match the new zero location.
          data[i+1] = 20-i; // Put an offset to the final byte into this location for now.
          lastzero = i+1;
        }
      else data[i+1] = packet[i];
    }

    Serial.write(data, 22);
  }
}

/*******************************************************************************************************
 * 
 * Command Handlers. Generally USB responses are only sent if debugSerial is true. The exception is
 * the SetDebug command, which always responds since if they're using it they probably need that info.
 * 
 * Every single distinct serial command has an associated handler below which is automatically called
 * if the serial command is received. In each case, numArgs is used to guarantee that the general
 * command format is valid, the intArg and floatArg functions pull in arguments from the serial command
 * while doing typechecking (it returns with an error if the data types are mismatched), and data
 * validation is immediately performed to test that the arguments are valid for our command.
 * 
 * Some commands are goofy, for legacy reasons. They are emulating the X2 firmware (version Stim24ChX72)
 * to maintain compatibility with old control scripts and old versions of the LabView software. New
 * commands are introduced to allow access to new functionality not present on the X2 board.
 * 
 *******************************************************************************************************/

serialcommand(getVersion) {
  Serial.println(String(Title) + " Version: " + String(FirmwareVersion));
}

serialcommand(startStreaming) {
  streaming = true;
}

serialcommand(stopStreaming) {
  streaming = false;
}

#define chunkSize 8000

serialcommand(adcRead) {  
  unsigned int analog[8];
  uint8_t datapacket[chunkSize*22];
  
  for (unsigned int samplenum=0; samplenum<chunkSize; samplenum++) {
    unsigned long currenttime = micros();
    
    // Get the analog input values.
    ADC::Sync_result result = adc->analogSyncRead(A1, A2);
    analog[6] = result.result_adc0;
    analog[7] = result.result_adc1;
    result = adc->analogSyncRead(A4, A3);
    analog[5] = result.result_adc0;
    analog[4] = result.result_adc1;
    analog[2] = adc->analogRead(A5);
    analog[3] = adc->analogRead(A6);
    analog[0] = adc->analogRead(A7);
    analog[1] = adc->analogRead(A8);
    
    byte packet[22]; // 16 bytes of data, four bytes for timestamp, two bytes of overhead in COBS format.

    packet[0] = currenttime>>24;
    packet[1] = currenttime>>16;
    packet[2] = currenttime>>8;
    packet[3] = currenttime;
    packet[4] = analog[0]>>8;
    packet[5] = analog[0];
    packet[6] = analog[1]>>8;
    packet[7] = analog[1];
    packet[8] = analog[2]>>8;
    packet[9] = analog[2];
    packet[10] = analog[3]>>8;
    packet[11] = analog[3];
    packet[12] = analog[4]>>8;
    packet[13] = analog[4];
    packet[14] = analog[5]>>8;
    packet[15] = analog[5];
    packet[16] = analog[6]>>8;
    packet[17] = analog[6];
    packet[18] = analog[7]>>8;
    packet[19] = analog[7];

    // Convert to COBS format.
    datapacket[samplenum*22] = 21; // Default to the end of the packet.
    datapacket[21+samplenum*22] = 0; // This is always set to the termination character.
    unsigned int lastzero = 0;
    
    for (unsigned int i=0; i<20; i++) {
      if (packet[i] == 0) {
          datapacket[lastzero+samplenum*22] = i+1-lastzero; // Change the last offset byte to match the new zero location.
          datapacket[i+1+samplenum*22] = 20-i; // Put an offset to the final byte into this location for now.
          lastzero = i+1;
        }
      else datapacket[i+1+samplenum*22] = packet[i];
    }
  }
  
  Serial.write(datapacket, chunkSize*22);
}

